
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxDataSources.pas.                                    }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxDataSources;

{$include Jvx.inc}

interface

uses
  Classes, DB, DBCtrls, TypInfo, JvxClasses, JvxUtils;

type

  { TJvxObjectDataField }

  TJvxObjectDataField = class(TJvxDataField)
  protected
    function GetInstance: TObject; virtual; abstract;
    procedure SetInstance(Value: TObject); virtual; abstract;
    property Instance: TObject read GetInstance write SetInstance;
  end;

  { TJvxObjectDataSource }

  TJvxObjectDataSource = class(TJvxDataSource)
  private
    FCurrentInstance: TObject;
    procedure SetCurrentInstance(Value: TObject);
  protected
    procedure ValidateField(Field: TJvxDataField); override;
    property CurrentInstance: TObject read FCurrentInstance write SetCurrentInstance;
  end;

  { TJvxDBDataField }

  // todo -crearchitecture -oosych: make TJvxDBDataField descend from TJvxObjectDataField
  TJvxDBDataField = class(TJvxDataField)
  private
    FDataLink: TFieldDataLink;
    procedure DataChanged(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    class function FieldTypeToTypeInfo(DataType: TFieldType): PTypeInfo;
  protected
    function GetTypeInfo: PTypeInfo; override;
    function InternalGetAsFloat: Extended; override;
    function InternalGetAsOrdinal: TJvxOrdinal; override;
    function InternalGetAssigned: Boolean; override;
    function InternalGetAsString: string; override;
    function InternalGetAsVariant: Variant; override;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property OnChange;
  end;

  { TJvxDBDataSource }

  TJvxDBDataSource = class(TJvxDataSource)
  private
    FDataLink: TDataLink;
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    function GetDataSet: TDataSet;
  protected
    procedure CreateDefaultFields; override;
    function InternalCount: Integer; override;
    function InternalGetNode(Index: Integer;
      Prototype: TJvxNode): TJvxNode; override;
    procedure InternalOpen; override;
    procedure ValidateField(Field: TJvxDataField); override;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    property DataSet: TDataSet read GetDataSet;
  published
    property DataSource: TDataSource read GetDataSource write SetDataSource;
  end;

  { TJvxPropertyDataField }

  // todo -crearchitecture -oosych: make TJvxPropertyDataField descend from TJvxObjectDataField
  TJvxPropertyDataField = class(TJvxDataField)
  private
    FInstance: TObject;
    FPropertyName: string;
    FPropInfo: PPropInfo;
    class function GetPropertyInfo(Instance: TObject; PropName: string): PPropInfo;
    function GetTypeData: PTypeData;
    procedure SetInstance(Value: TObject);
    procedure SetPropertyName(const Value: string);
  protected
    function GetTypeInfo: PTypeInfo; override;
    function InternalGetAsFloat: Extended; override;
    function InternalGetAsOrdinal: TJvxOrdinal; override;
    function InternalGetAssigned: Boolean; override;
    function InternalGetAsString: string; override;
    function InternalGetAsVariant: Variant; override;

    property PropInfo: PPropInfo read FPropInfo;
    property TypeInfo: PTypeInfo read GetTypeInfo;
    property TypeData: PTypeData read GetTypeData;
  published
    property Instance: TObject read FInstance write SetInstance;
    property OnChange;
    property PropertyName: string read FPropertyName write SetPropertyName;
  end;

  { TJvxFieldsDataSource }

  // todo -crearchitecture -oosych: Make TJvxFieldsDataSource descend from TJvxObjectDataSource
  TJvxFieldsDataSource = class(TJvxDataSource)
  private
    FDataSet: TDataSet;
    procedure SetDataSet(Value: TDataSet);
  protected
    procedure CreateDefaultFields; override;
    function InternalCount: Integer; override;
    function InternalGetNode(Index: Integer;
      Prototype: TJvxNode): TJvxNode; override;
    procedure InternalOpen; override;
    procedure Notification(Component: TComponent;
      Operation: TOperation); override;
    procedure ValidateField(Field: TJvxDataField); override;
  published
    property DataSet: TDataSet read FDataSet write SetDataSet;
  end;

implementation

uses
  SysUtils;

resourcestring
  SInvalidPropertyType = 'Property %s has unsupported type';
  SMissingDataSetProperty = '%s. Missing DataSet property';
  SMissingDataSourceProperty = '%s. Missing DataSource property';
  SUnknownProperty = 'Property %s does not exist';

const
  SupportedObjectProperties = [tkInteger, tkChar, tkEnumeration, tkFloat,
    tkString, tkLString, tkVariant];

{ TJvxObjectDataSource }

procedure TJvxObjectDataSource.SetCurrentInstance(Value: TObject);
var
  I: Integer;
begin
  if FCurrentInstance <> Value then
  begin
    FCurrentInstance := Value;
    for I := 0 to Fields.Count - 1 do
      (Fields[I] as TJvxObjectDataField).Instance := Value;
  end;
end;

procedure TJvxObjectDataSource.ValidateField(Field: TJvxDataField);
begin
  if Field is TJvxObjectDataField then Exit;
  inherited;
end;

{ TJvxDBDataField }

constructor TJvxDBDataField.Create(Owner: TComponent);
begin
  inherited;
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChanged;
end;

procedure TJvxDBDataField.DataChanged(Sender: TObject);
begin
  Changed;
end;

destructor TJvxDBDataField.Destroy;
begin
  FDataLink.Free;
  inherited;
end;

class function TJvxDBDataField.FieldTypeToTypeInfo(DataType: TFieldType): PTypeInfo;
begin
  case DataType of
    ftString:     Result := System.TypeInfo(string);
    ftSmallint:   Result := System.TypeInfo(SmallInt);
    ftInteger:    Result := System.TypeInfo(Integer);
    ftWord:       Result := System.TypeInfo(Word);
    ftBoolean:    Result := System.TypeInfo(Boolean);
    ftFloat:      Result := System.TypeInfo(Extended);
    ftCurrency:   Result := System.TypeInfo(Currency);
    ftBCD:        Result := System.TypeInfo(Currency);
    ftDate:       Result := System.TypeInfo(TDateTime);
    ftTime:       Result := System.TypeInfo(TDateTime);
    ftDateTime:   Result := System.TypeInfo(TDateTime);
    ftAutoInc:    Result := System.TypeInfo(Integer);
    ftMemo:       Result := System.TypeInfo(string);
    ftFmtMemo:    Result := System.TypeInfo(string);
    ftFixedChar:  Result := System.TypeInfo(string);
    ftWideString: Result := System.TypeInfo(string);
    ftLargeint:   Result := System.TypeInfo(LargeInt);
    ftVariant:    Result := System.TypeInfo(Variant);
  else
    Result := nil;
  end;
end;

function TJvxDBDataField.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

function TJvxDBDataField.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

function TJvxDBDataField.GetField: TField;
begin
  Result := FDataLink.Field;
end;

function TJvxDBDataField.GetTypeInfo: PTypeInfo;
begin
  if Field <> nil then
    Result := FieldTypeToTypeInfo(Field.DataType)
  else
    Result := inherited GetTypeInfo;
end;

function TJvxDBDataField.InternalGetAsFloat: Extended;
begin
  Result := Field.AsFloat;
end;

function TJvxDBDataField.InternalGetAsOrdinal: TJvxOrdinal;
begin
  if Field.DataType = ftBoolean then
    Result := Ord(Field.AsBoolean)
  else
    Result := Field.AsInteger;
end;

function TJvxDBDataField.InternalGetAssigned: Boolean;
begin
  Result := (Field <> nil) and (not Field.IsNull);
end;

function TJvxDBDataField.InternalGetAsString: string;
begin
  Result := Field.AsString;
end;

function TJvxDBDataField.InternalGetAsVariant: Variant;
begin
  Result := Field.AsVariant;
end;

procedure TJvxDBDataField.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

procedure TJvxDBDataField.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

{ TJvxDBDataSource }

constructor TJvxDBDataSource.Create(Owner: TComponent);
begin
  inherited;
  FDataLink := TDataLink.Create;
end;

procedure TJvxDBDataSource.CreateDefaultFields;
var
  I: Integer;
  Field: TJvxDBDataField;
begin
  for I := 0 to DataSet.Fields.Count - 1 do
    if TJvxDBDataField.FieldTypeToTypeInfo(DataSet.Fields[I].DataType) <> nil then
    begin
      Field := TJvxDBDataField.Create(Self);
      try
        Field.DataSource := DataSource;
        Field.FieldName := DataSet.Fields[I].FieldName;
        Field.DataField := DataSet.Fields[I].FieldName;
        Fields.Add(Field);
      except
        Field.Free;
        raise;
      end;
    end;
end;

destructor TJvxDBDataSource.Destroy;
begin
  FDataLink.Free;
  inherited;
end;

function TJvxDBDataSource.GetDataSet: TDataSet;
begin
  Result := FDataLink.DataSet;
end;

function TJvxDBDataSource.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

function TJvxDBDataSource.InternalCount: Integer;
begin
  if DataSet <> nil then
    Result := DataSet.RecordCount
  else
    Result := inherited InternalCount;
end;

function TJvxDBDataSource.InternalGetNode(Index: Integer;
  Prototype: TJvxNode): TJvxNode;
var
  Offset: Integer;
begin
  if DataSet <> nil then
  begin
    DataSet.DisableControls;
    try
      Offset := Index - (DataSet.RecNo - 1);
      if DataSet.MoveBy(Offset) = Offset then
      begin
        Result := Prototype;
        Exit;
      end;
    finally
      DataSet.EnableControls;
    end;
  end;
  Result := inherited InternalGetNode(Index, Prototype);
end;

procedure TJvxDBDataSource.InternalOpen;
begin
  if DataSource = nil then
    raise EJvxError.CreateFmt(SMissingDataSourceProperty, [Name]);
  if DataSet = nil then
    raise EJvxError.CreateFmt(SMissingDataSetProperty, [DataSource.Name]);
end;

procedure TJvxDBDataSource.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

procedure TJvxDBDataSource.ValidateField(Field: TJvxDataField);
begin
  if Field is TJvxDBDataField then Exit;
  inherited;
end;

{ TJvxPropertyDataField }

class function TJvxPropertyDataField.GetPropertyInfo(Instance: TObject;
  PropName: string): PPropInfo;
begin
  if (Instance <> nil) and (PropName <> EmptyStr) then
  begin
    Result := GetPropInfo(Instance, PropName);
    if Result = nil then
      raise EJvxError.CreateFmt(SUnknownProperty, [PropName]);
    if not (Result.PropType^.Kind in SupportedObjectProperties) then
      raise EJvxError.CreateFmt(SInvalidPropertyType, [PropName]);
    Exit;
  end;
  Result := nil;
end;

function TJvxPropertyDataField.GetTypeData: PTypeData;
begin
  if TypeInfo <> nil then
    Result := TypInfo.GetTypeData(TypeInfo)
  else
    Result := nil;
end;

function TJvxPropertyDataField.GetTypeInfo: PTypeInfo;
begin
  if (PropInfo <> nil) and (PropInfo.PropType <> nil) then
    Result := PropInfo.PropType^
  else
    Result := inherited GetTypeInfo;
end;

function TJvxPropertyDataField.InternalGetAsFloat: Extended;
begin
  Result := GetFloatProp(FInstance, PropInfo);
end;

function TJvxPropertyDataField.InternalGetAsOrdinal: TJvxOrdinal;
begin
  Result := GetOrdProp(FInstance, PropInfo);
end;

function TJvxPropertyDataField.InternalGetAssigned: Boolean;
begin
  if PropInfo = nil then
  begin
    Result := False;
    Exit;
  end;

  case TypeInfo.Kind of
    tkInteger, tkChar, tkEnumeration, tkFloat:
      Result := True;
    tkString, tkLString:
      Result := GetStrProp(Instance, PropInfo) <> EmptyStr;
    tkVariant:
      Result := not VarIsEmpty(GetVariantProp(Instance, PropInfo)) and
        not VarIsNull(GetVariantProp(Instance, PropInfo));
  else
    Result := False;
  end;
end;

function TJvxPropertyDataField.InternalGetAsString: string;
begin
  Result := GetStrProp(Instance, PropInfo);
end;

function TJvxPropertyDataField.InternalGetAsVariant: Variant;
begin
  Result := GetVariantProp(Instance, PropInfo);
end;

procedure TJvxPropertyDataField.SetInstance(Value: TObject);
begin
  if FInstance <> Value then
  begin
    FInstance := Value;
    FPropInfo := GetPropertyInfo(FInstance, PropertyName);
    Changed;
  end;
end;

procedure TJvxPropertyDataField.SetPropertyName(const Value: string);
begin
  if FPropertyName <> Value then
  begin
    FPropertyName := Value;
    FPropInfo := GetPropertyInfo(FInstance, PropertyName);
    Changed;
  end;
end;

{ TJvxFieldsDataSource }

procedure TJvxFieldsDataSource.CreateDefaultFields;
var
  I: Integer;
  Field: TJvxPropertyDataField;
  PropCount: Integer;
  PropInfo: PPropInfo;
  PropList: PPropList;
begin
  PropCount := GetPropList(TField.ClassInfo, SupportedObjectProperties, nil);
  GetMem(PropList, PropCount * SizeOf(PPropInfo));
  try
    GetPropList(TField.ClassInfo, SupportedObjectProperties, PropList);
    for I := 0 to PropCount - 1 do
    begin
      PropInfo := PropList^[I];
      Field := TJvxPropertyDataField.Create(Self);
      try
        Field.FieldName    := PropInfo.Name;
        Field.PropertyName := PropInfo.Name;
        Fields.Add(Field);
      except
        Field.Free;
        raise;
      end;
    end;
  finally
    FreeMem(PropList);
  end;
end;

function TJvxFieldsDataSource.InternalCount: Integer;
begin
  Result := DataSet.Fields.Count;
end;

function TJvxFieldsDataSource.InternalGetNode(Index: Integer;
  Prototype: TJvxNode): TJvxNode;
var
  DatabaseField: TField;
  I: Integer;
begin
  DatabaseField := DataSet.Fields[Index];
  for I := 0 to Fields.Count - 1 do
    TJvxPropertyDataField(Fields[I]).Instance := DatabaseField;
  Result := Prototype;
end;

procedure TJvxFieldsDataSource.InternalOpen;
begin
  if DataSet = nil then
    raise EJvxError.CreateFmt(SMissingDataSetProperty, [Name]);
end;

procedure TJvxFieldsDataSource.Notification(Component: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (Component = FDataSet) then
    SetDataSet(nil);
end;

procedure TJvxFieldsDataSource.SetDataSet(Value: TDataSet);
begin
  if FDataSet <> Value then
  begin
    if FDataSet <> nil then
      FDataSet.RemoveFreeNotification(Self);
    FDataSet := Value;
    if FDataSet <> nil then
      FDataSet.FreeNotification(Self);
  end;
end;

procedure TJvxFieldsDataSource.ValidateField(Field: TJvxDataField);
begin
  if Field is TJvxPropertyDataField then Exit;
  inherited;
end;

end.

