
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxMemoryManager.pas.                                  }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxMemoryManager;

{$include Jvx.inc}

{$ifdef JvxDebugMem}
  {$define __JvxDebugMemDeclarations__}
  {$define __JvxDebugMemInterface__}
  {$define __JvxDebugMemImplementation__}
{$endif}

{$ifdef JvxDebugMemShared}
  {$define JvxDebugMem}
  {$define __JvxDebugMemDeclarations__}
  {$define __JvxDebugMemInterface__}
{$endif}

{$ifdef JvxDebugMemDll}
  {$define __JvxDebugMemDeclarations__}
  {$define __JvxDebugMemImplementation__}
{$else}
  {$define __JvxDebugMemInterface__}
{$endif}

interface

type
  TJvxDebugHeapError =
  (
    heHeapOk,           // everything's fine
    heReleaseErr,       // operating system returned an error when we released
    heDecommitErr,      // operating system returned an error when we decommited
    heBadCommittedList, // list of committed blocks looks bad
    heBadFiller1,       // filler block is bad
    heBadFiller2,       // filler block is bad
    heBadFiller3,       // filler block is bad
    heBadCurAlloc,      // current allocation zone is bad
    heCantInit,         // couldn't initialize
    heBadUsedBlock,     // used block looks bad
    heBadPrevBlock,     // prev block before a used block is bad
    heBadNextBlock,     // next block after a used block is bad
    heBadFreeList,      // free list is bad
    heBadFreeBlock,     // free block is bad
    heBadBalance        // free list doesn't correspond to blocks marked free
  );

const
  JvxGetMemGarbageChar = #$CC;
  JvxFreeMemGarbageChar = #$CD;
  JvxDebugMemDll = 'JvxDebugMem.dll';

{ Memory manager implementation functions }

{$ifdef __JvxDebugMemDeclarations__}
function JvxDebugGetMem(Size: Integer): Pointer;
function JvxDebugFreeMem(Ptr: Pointer): Integer;
function JvxDebugReallocMem(Ptr: Pointer; Size: Integer): Pointer;
{$endif}

{ Validation functions }

function JvxAlignedMemoryBlockSize(Size: Integer): Integer;
function JvxGetMemoryBlockSize(Ptr: Pointer; out BlockSize: Integer): Boolean;
function JvxDebugAllocMemCount: Integer;
function JvxDebugAllocMemSize: Integer;
function JvxDebugHeapError: TJvxDebugHeapError;
function JvxDebugHeapStatus: THeapStatus;
function JvxVerifyMemoryBlock(Ptr: Pointer): Boolean;

implementation

{$ifdef __JvxDebugMemImplementation__}

{ Include standard Delphi's memory manager implementation }

const
  kernel = 'kernel32.dll';

function SysGetMem(Size: Integer): Pointer; forward;
function SysFreeMem(P: Pointer): Integer; forward;
function SysReallocMem(P: Pointer; Size: Integer): Pointer; forward;

{$include GetMem.inc}

{ Validation functions }

function JvxAlignedMemoryBlockSize(Size: Integer): Integer;
begin
  Result := Size + SizeOf(TUsed) + (cAlign - 1);
  Result := Result and not (cAlign - 1);
  if Result < SizeOf(TFree) then
    Result := SizeOf(TFree);
  Result := Result - SizeOf(TUsed);
end;

function JvxGetMemoryBlockHeader(Ptr: Pointer; out BlockHeader: PUsed): Boolean;
begin
  Result := False;
  if Ptr = nil then Exit;
  BlockHeader := PUsed(PChar(Ptr) - SizeOf(TUsed));
  if (BlockHeader.SizeFlags and cThisUsedFlag) <> cThisUsedFlag then Exit;
  if (BlockHeader.SizeFlags and cFillerFlag) = cFillerFlag then Exit;
  Result := True;
end;

function JvxVerifyMemoryBlock(Ptr: Pointer): Boolean;

  function FindBlockDescriptor(Block: PUsed;
    out BlockDesc: PBlockDesc): Boolean;
  begin
    BlockDesc := CommittedRoot.Next;
    repeat
      if (Block >= BlockDesc.Addr) and
         (Block < (BlockDesc.Addr + BlockDesc.Size)) then
      begin
        Result := True;
        Exit;
      end;
      BlockDesc := BlockDesc.Next;
    until BlockDesc = @CommittedRoot;
    Result := False;
  end;

  function FindBlockInDescriptor(Block: PUsed; BlockDesc: PBlockDesc): Boolean;
  var
    CurBlock: PChar;
    EndAddr: PChar;
    BlockSize: Integer;
  begin
    CurBlock := BlockDesc.Addr;
    EndAddr := BlockDesc.Addr + BlockDesc.Size;

    // Search through the list of memory blocks in the descriptor
    while CurBlock < EndAddr do
    begin
      // If the block is found
      if PUsed(CurBlock) = Block then
      begin
        Result := True;
        Exit;
      end;

      // Determine size of the current block
      if (CurBlock = CurAlloc) and (RemBytes > 0) then
        BlockSize := RemBytes
      else if (PUsed(CurBlock).SizeFlags and cThisUsedFlag) <> cThisUsedFlag then
        BlockSize := PFree(CurBlock).Size
      else
        BlockSize := PUsed(CurBlock).SizeFlags and not cFlags;

      // If block points into the middle of allocated memory block
      if Block < (CurBlock + BlockSize) then
        Break;

      // Go to the next block in descriptor
      Inc(CurBlock, BlockSize);
    end;
    Result := False;
  end;

var
  Block: PUsed;
  BlockDesc: PBlockDesc;
begin
  if IsMultiThread then
    EnterCriticalSection(HeapLock);
  try
    Result := False;
    if not JvxGetMemoryBlockHeader(Ptr, Block) then Exit;
    if PChar(Block) = CurAlloc then Exit;
    if not FindBlockDescriptor(Block, BlockDesc) then Exit;
    if not FindBlockInDescriptor(Block, BlockDesc) then Exit;
    Result := True;
  finally
    if IsMultiThread then
      LeaveCriticalSection(HeapLock);
  end;
end;

function JvxGetMemoryBlockSize(Ptr: Pointer; out BlockSize: Integer): Boolean;
var
  BlockHeader: PUsed;
begin
  Result := False;
  if JvxGetMemoryBlockHeader(Ptr, BlockHeader) then
  begin
    BlockSize := (BlockHeader.SizeFlags and not cFlags) - SizeOf(TUsed);
    Result := True;
  end;
end;

{ Memory manager implementation functions }

function JvxDebugGetMem(Size: Integer): Pointer;
var
  BlockSize: Integer;
begin
  if IsMultiThread then
    EnterCriticalSection(HeapLock);
  try
    // Call the standard implementation
    Result := SysGetMem(Size);

    // Fill allocated memory with garbage
    if JvxGetMemoryBlockSize(Result, BlockSize) then
      FillChar(Result^, BlockSize, JvxGetMemGarbageChar);
  finally
    if IsMultiThread then
      LeaveCriticalSection(HeapLock);
  end;
end;

function JvxDebugFreeMem(Ptr: Pointer): Integer;
var
  BlockSize: Integer;
begin
  if IsMultiThread then
    EnterCriticalSection(HeapLock);
  try
    // Fill freed memory with garbage
    if JvxGetMemoryBlockSize(Ptr, BlockSize) then
      FillChar(Ptr^, BlockSize, JvxFreeMemGarbageChar);

    // Call the standard implementation
    Result := SysFreeMem(Ptr);
  finally
    if IsMultiThread then
      LeaveCriticalSection(HeapLock);
  end;
end;

function JvxDebugReallocMem(Ptr: Pointer; Size: Integer): Pointer;
var
  BlockSize: Integer;
begin
  if IsMultiThread then
    EnterCriticalSection(HeapLock);
  try
    if JvxGetMemoryBlockSize(Ptr, BlockSize) then
    begin
      // Allocate new memory block
      Result := JvxDebugGetMem(Size);
      if Result = nil then
        Exit;

      // Move the block to the new location
      if BlockSize < Size then
        Move(Ptr^, Result^, BlockSize)
      else
        Move(Ptr^, Result^, Size);

      // Free old memory block
      if JvxDebugFreeMem(Ptr) <> cHeapOk then
      begin
        JvxDebugFreeMem(Result);
        Result := nil;
        Exit;
      end;

      Ptr := Result;
    end;

    // Always call the standard implementation
    Result := SysReallocMem(Ptr, Size);
  finally
    if IsMultiThread then
      LeaveCriticalSection(HeapLock);
  end;
end;

function JvxDebugHeapStatus: THeapStatus;
begin
  Result := GetHeapStatus;
end;

function JvxDebugHeapError: TJvxDebugHeapError;
begin
  Result := TJvxDebugHeapError(HeapErrorCode);
end;

function JvxDebugAllocMemCount: Integer;
begin
  Result := AllocMemCount;
end;

function JvxDebugAllocMemSize: Integer;
begin
  Result := AllocMemSize;
end;

{$endif}

{$ifdef __JvxDebugMemInterface__}

{$ifdef JvxDebugMemShared}

function JvxAlignedMemoryBlockSize(Size: Integer): Integer; external JvxDebugMemDll;
function JvxGetMemoryBlockSize(Ptr: Pointer; out BlockSize: Integer): Boolean; external JvxDebugMemDll;
function JvxDebugAllocMemCount: Integer; external JvxDebugMemDll;
function JvxDebugAllocMemSize: Integer; external JvxDebugMemDll;
function JvxDebugGetMem(Size: Integer): Pointer; external JvxDebugMemDll;
function JvxDebugHeapError: TJvxDebugHeapError; external JvxDebugMemDll;
function JvxDebugHeapStatus: THeapStatus; external JvxDebugMemDll;
function JvxDebugFreeMem(Ptr: Pointer): Integer; external JvxDebugMemDll;
function JvxDebugReallocMem(Ptr: Pointer; Size: Integer): Pointer; external JvxDebugMemDll;
function JvxVerifyMemoryBlock(Ptr: Pointer): Boolean; external JvxDebugMemDll;

{$endif}

{$ifndef JvxDebugMem}

function JvxAlignedMemoryBlockSize(Size: Integer): Integer;
begin
  Result := -1;
end;

function JvxGetMemoryBlockSize(Ptr: Pointer; out BlockSize: Integer): Boolean;
begin
  BlockSize := -1;
  Result := False;
end;

function JvxDebugHeapError: TJvxDebugHeapError;
begin
  Result := heHeapOk;
end;

function JvxDebugHeapStatus: THeapStatus;
begin
  Result := System.GetHeapStatus;
end;

function JvxDebugAllocMemCount: Integer;
begin
  Result := System.AllocMemCount;
end;

function JvxDebugAllocMemSize: Integer;
begin
  Result := System.AllocMemSize;
end;

function JvxVerifyMemoryBlock(Ptr: Pointer): Boolean;
begin
  Result := False;
end;

{$endif}

{$ifdef JvxDebugMem}

procedure InstallMemoryManager;
var
  MemoryManager: TMemoryManager;
begin
  MemoryManager.GetMem := JvxDebugGetMem;
  MemoryManager.FreeMem := JvxDebugFreeMem;
  MemoryManager.ReallocMem := JvxDebugReallocMem;
  SetMemoryManager(MemoryManager);
end;

procedure UninstallMemoryManager;
var
  MemoryManager: TMemoryManager;
begin
  MemoryManager.GetMem := System.SysGetMem;
  MemoryManager.FreeMem := System.SysFreeMem;
  MemoryManager.ReallocMem := System.SysReallocMem;
  SetMemoryManager(MemoryManager);
end;

{$endif}

{$endif}

initialization
{$ifdef __JvxDebugMemImplementation__}
  InitAllocator;
{$endif}
{$ifdef JvxDebugMem}
  if not System.IsMemoryManagerSet then
    InstallMemoryManager;
{$endif}

finalization
{$ifdef JvxDebugMem}
  if System.IsMemoryManagerSet then
    UninstallMemoryManager;
{$endif}
{$ifdef __JvxDebugMemImplementation__}
  UninitAllocator;
{$endif}

end.
