
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxTestClasses.pas.                                    }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxTestClasses;

{$include Jvx.inc}

interface

uses
  SysUtils, Classes, Contnrs;

type

  { TJvxTest }

  TJvxTestResults = class;
  TJvxTestClass = class of TJvxTest;
  TExceptionClass = class of Exception;
  TJvxTest = class(TComponent)
  private
    FMemorySize: Integer;
    FName: string;
    FTestResults: TJvxTestResults;
  protected
    procedure Assert(Expected: TExceptionClass); overload;
    procedure Assert(Expr: Boolean; const Msg: string = ''); overload;
    function GetTestCount: Integer; virtual; abstract;
    property MemorySize: Integer read FMemorySize;
    property TestResults: TJvxTestResults read FTestResults;
  public
    constructor Create(Owner: TComponent); overload; override;
    procedure Run(TestResults: TJvxTestResults); virtual;
    property Name: string read FName write FName;
    property TestCount: Integer read GetTestCount;
  end;

  { TJvxTestCase }

  TJvxTestMethod = procedure of object;
  TJvxTestCase = class(TJvxTest)
  private
    FTestMethod: TJvxTestMethod;
  protected
    function GetTestCount: Integer; override;
  public
    constructor Create(Owner: TComponent; TestMethod: TJvxTestMethod); overload;
    procedure Run(TestResults: TJvxTestResults); override;
    property TestCount: Integer read GetTestCount;
  end;

  { TJvxTestSuite }

  TJvxTestSuite = class(TJvxTest)
  private
    FTests: TObjectList;
  protected
    procedure CreateFixture;
    procedure DestroyFixture;
    function GetTestCount: Integer; override;
    procedure InternalCreateFixture; virtual;
    procedure InternalDestroyFixture; virtual;
    procedure RunTest(Test: TJvxTest; TestResults: TJvxTestResults);
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;

    procedure AddTest(Test: TJvxTest); overload;
    procedure AddTest(TestMethod: TJvxTestMethod); overload;
    procedure Run(TestResults: TJvxTestResults); override;
  end;

  { TJvxTestResults }

  TJvxTestResult = class;
  TJvxTestFailure = class;
  TJvxTestNotifyEvent = procedure (Test: TJvxTest) of object;
  TJvxTestFailureEvent = procedure (Failure: TJvxTestFailure) of object;
  TJvxTestResults = class(TObject)
  private
    FAborted: Boolean;
    FBreakOnFailure: Boolean;
    FErrors: TObjectList;
    FFailures: TObjectList;
    FResults: TObjectList;
    FOnTestError: TJvxTestFailureEvent;
    FOnTestFailure: TJvxTestFailureEvent;
    FOnTestStarted: TJvxTestNotifyEvent;
    FOnTestStopped: TJvxTestNotifyEvent;
    function GetError(Index: Integer): TJvxTestFailure;
    function GetFailure(Index: Integer): TJvxTestFailure;
    function GetErrorCount: Integer;
    function GetFailureCount: Integer;
    function GetResultCount: Integer;
    function GetTestResult(Index: Integer): TJvxTestResult;
  protected
    procedure AddError(Test: TJvxTest; E: Exception);
    procedure AddFailure(Test: TJvxTest; E: EAssertionFailed);
    procedure TestStarted(Test: TJvxTest);
    procedure TestStopped(Test: TJvxTest);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Abort;
    procedure Clear;
    property Aborted: Boolean read FAborted;
    property BreakOnFailure: Boolean read FBreakOnFailure write FBreakOnFailure;
    property ErrorCount: Integer read GetErrorCount;
    property Errors[Index: Integer]: TJvxTestFailure read GetError;
    property FailureCount: Integer read GetFailureCount;
    property Failures[Index: Integer]: TJvxTestFailure read GetFailure;
    property ResultCount: Integer read GetResultCount;
    property Results[Index: Integer]: TJvxTestResult read GetTestResult;
    property OnTestError: TJvxTestFailureEvent read FOnTestError write FOnTestError;
    property OnTestFailure: TJvxTestFailureEvent read FOnTestFailure write FOnTestFailure;
    property OnTestStarted: TJvxTestNotifyEvent read FOnTestStarted write FOnTestStarted;
    property OnTestStopped: TJvxTestNotifyEvent read FOnTestStopped write FOnTestStopped;
  end;

  { TJvxTestResult }

  TJvxTestResult = class(TObject)
  private
    FMessage: string;
    FTestName: string;
  public
    constructor Create(Test: TJvxTest; Message: string);
    property Message: string read FMessage;
    property TestName: string read FTestName;
  end;

  { TJvxTestFailure }

  TJvxTestFailure = class(TJvxTestResult)
  private
    FErrorClass: TExceptionClass;
  public
    constructor Create(Test: TJvxTest; Error: Exception);
    property ErrorClass: TExceptionClass read FErrorClass;
  end;

{ Test Registry }

procedure JvxRegisterTestClass(TestClass: TJvxTestClass);
function JvxFindTestClass(const ClassName: string): TJvxTestClass;
function JvxGetTestClass(const ClassName: string): TJvxTestClass;
procedure JvxGetTestClassNames(ClassNames: TStrings);

implementation

uses
  Windows, Consts, JvxMemoryManager;

resourcestring
  STestStarted = 'Test started.';
  STestStopped = 'Test stopped.';
  SExceptionExpected = 'Exception %s was expected.';
  SUnexpectedException = 'Exception %s was expected. Unexpected exception found: %s - %s. ';

{ TJvxTest }

procedure TJvxTest.Assert(Expected: TExceptionClass);
begin
  if ExceptObject = nil then
    Assert(False, Format(SExceptionExpected, [Expected.ClassName]))
  else if ExceptObject is EAssertionFailed then
    raise EAssertionFailed.Create(EAssertionFailed(ExceptObject).Message)
  else
    Assert(ExceptObject is Expected, Format(SUnexpectedException,
      [Expected.ClassName, ExceptObject.ClassName,
        (ExceptObject as Exception).Message]));
end;

procedure TJvxTest.Assert(Expr: Boolean; const Msg: string);
begin
  if (Expr = False) and TestResults.BreakOnFailure then
    DebugBreak;
  System.Assert(Expr, Msg);
end;

constructor TJvxTest.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FName := ClassName;
end;

procedure TJvxTest.Run(TestResults: TJvxTestResults);
begin
  FTestResults := TestResults;
end;

{ TJvxTestCase }

constructor TJvxTestCase.Create(Owner: TComponent; TestMethod: TJvxTestMethod);
begin
  inherited Create(Owner);
  FTestMethod := TestMethod;
  with TMethod(FTestMethod) do
  begin
    FName := TObject(Data).ClassName;
    FName := FName + '.' + TObject(Data).MethodName(Code);
  end;
end;

function TJvxTestCase.GetTestCount: Integer;
begin
  Result := 1;
end;

procedure TJvxTestCase.Run(TestResults: TJvxTestResults);
var
  SavedMemorySize: Integer;
begin
  inherited;
  SavedMemorySize := JvxDebugAllocMemSize;
  try
    if Assigned(FTestMethod) then
      FTestMethod;
  finally
    FMemorySize := JvxDebugAllocMemSize - SavedMemorySize;
  end;
end;

{ TJvxTestSuite }

procedure TJvxTestSuite.AddTest(Test: TJvxTest);
begin
  FTests.Add(Test);
end;

procedure TJvxTestSuite.AddTest(TestMethod: TJvxTestMethod);
begin
  AddTest(TJvxTestCase.Create(Self, TestMethod));
end;

constructor TJvxTestSuite.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FTests := TObjectList.Create(True);
end;

procedure TJvxTestSuite.CreateFixture;
var
  SavedMemorySize: Integer;
begin
  SavedMemorySize := JvxDebugAllocMemSize;
  InternalCreateFixture;
  FMemorySize := FMemorySize + JvxDebugAllocMemSize - SavedMemorySize;
end;

destructor TJvxTestSuite.Destroy;
begin
  FTests.Free;
  inherited;
end;

procedure TJvxTestSuite.DestroyFixture;
var
  SavedMemorySize: Integer;
begin
  SavedMemorySize := JvxDebugAllocMemSize;
  InternalDestroyFixture;
  FMemorySize := FMemorySize + JvxDebugAllocMemSize - SavedMemorySize;
end;

function TJvxTestSuite.GetTestCount: Integer;
var
  I: Integer;
begin
  Result := 1;
  for I := 0 to FTests.Count - 1 do
    Result := Result + TJvxTest(FTests[I]).TestCount;
end;

procedure TJvxTestSuite.InternalCreateFixture;
begin
end;

procedure TJvxTestSuite.InternalDestroyFixture;
begin
end;

procedure TJvxTestSuite.Run(TestResults: TJvxTestResults);
var
  I: Integer;
  Test: TJvxTest;
begin
  inherited;
  TestResults.TestStarted(Self);
  try
    for I := 0 to FTests.Count - 1 do
    begin
      if TestResults.Aborted then
        Exit;

      try
        Test := TJvxTest(FTests[I]);
        try
          FMemorySize := 0;
          try
            CreateFixture;
            RunTest(Test, TestResults);
          finally
            DestroyFixture;
          end;
          Assert(FMemorySize <= 0, Format('Memory leak of %d bytes', [FMemorySize]));
        except
          on E: Exception do
            raise ExceptClass(E.ClassType).Create(Test.Name + '. ' + E.Message);
        end;
      except
        on E: EAssertionFailed do
          TestResults.AddFailure(Self, E);
        on E: Exception do
          TestResults.AddError(Self, E);
      end;
    end;
  finally
    TestResults.TestStopped(Self);
  end;
end;

procedure TJvxTestSuite.RunTest(Test: TJvxTest; TestResults: TJvxTestResults);
begin
  TestResults.TestStarted(Test);
  try
    if TestResults.Aborted then
      Exit;
    try
      Test.Run(TestResults);
      FMemorySize := FMemorySize + Test.MemorySize;
    except
      on E: EAssertionFailed do
        TestResults.AddFailure(Test, E);
      on E: Exception do
        TestResults.AddError(Test, E);
    end;
  finally
    TestResults.TestStopped(Test);
  end;
end;

{ TJvxTestResults }

procedure TJvxTestResults.Abort;
begin
  FAborted := True;
end;

procedure TJvxTestResults.AddError(Test: TJvxTest; E: Exception);
var
  Error: TJvxTestFailure;
begin
  Error := TJvxTestFailure.Create(Test, E);
  try
    FResults.Add(Error);
    FErrors.Add(Error);
    if Assigned(FOnTestError) then
      FOnTestError(Error);
  except
    Error.Free;
    raise;
  end;
end;

procedure TJvxTestResults.AddFailure(Test: TJvxTest; E: EAssertionFailed);
var
  Failure: TJvxTestFailure;
begin
  Failure := TJvxTestFailure.Create(Test, E);
  try
    FResults.Add(Failure);
    FFailures.Add(Failure);
    if Assigned(FOnTestFailure) then
      FOnTestFailure(Failure);
  except
    Failure.Free;
    raise;
  end;
end;

procedure TJvxTestResults.Clear;
begin
  FAborted := false;
  FErrors.Clear;
  FFailures.Clear;
  FResults.Clear;
end;

constructor TJvxTestResults.Create;
begin
  inherited;
  FErrors := TObjectList.Create(False);
  FFailures := TObjectList.Create(False);
  FResults := TObjectList.Create(True);
end;

destructor TJvxTestResults.Destroy;
begin
  FErrors.Free;
  FFailures.Free;
  FResults.Free;
  inherited;
end;

function TJvxTestResults.GetError(Index: Integer): TJvxTestFailure;
begin
  Result := FErrors[Index] as TJvxTestFailure;
end;

function TJvxTestResults.GetErrorCount: Integer;
begin
  Result := FErrors.Count;
end;

function TJvxTestResults.GetFailure(Index: Integer): TJvxTestFailure;
begin
  Result := FFailures[Index] as TJvxTestFailure;
end;

function TJvxTestResults.GetFailureCount: Integer;
begin
  Result := FFailures.Count;
end;

function TJvxTestResults.GetResultCount: Integer;
begin
  Result := FResults.Count;
end;

function TJvxTestResults.GetTestResult(Index: Integer): TJvxTestResult;
begin
  Result := FResults[Index] as TJvxTestResult;
end;

procedure TJvxTestResults.TestStarted(Test: TJvxTest);
begin
  if Assigned(FOnTestStarted) then
    FOnTestStarted(Test);
end;

procedure TJvxTestResults.TestStopped(Test: TJvxTest);
begin
  if Assigned(FOnTestStopped) then
    FOnTestStopped(Test);
end;

{ TJvxTestResult }

constructor TJvxTestResult.Create(Test: TJvxTest; Message: string);
begin
  inherited Create;
  FTestName := Test.Name;
  FMessage := Message;
end;

{ TJvxTestFailure }

constructor TJvxTestFailure.Create(Test: TJvxTest; Error: Exception);
begin
  inherited Create(Test, Error.Message);
  FErrorClass := TExceptionClass(Error.ClassType);
end;

{ Test Registry }

var
  TestRegistry: TClassList;

procedure JvxRegisterTestClass(TestClass: TJvxTestClass);
begin
  if TestRegistry.IndexOf(TestClass) >= 0 then
    raise Exception.CreateFmt(SDuplicateClass, [TestClass.ClassName]);
  TestRegistry.Add(TestClass);
end;

function JvxFindTestClass(const ClassName: string): TJvxTestClass;
var
  I: Integer;
begin
  for I := 0 to TestRegistry.Count - 1 do
  begin
    Result := TJvxTestClass(TestRegistry[I]);
    if Result.ClassNameIs(ClassName) then
      Exit;
  end;
  Result := nil;
end;

function JvxGetTestClass(const ClassName: string): TJvxTestClass;
begin
  Result := JvxFindTestClass(ClassName);
  if Result = nil then
    raise EClassNotFound.CreateFmt(SClassNotFound, [ClassName]);
end;

procedure JvxGetTestClassNames(ClassNames: TStrings);
var
  I: Integer;
begin
  ClassNames.BeginUpdate;
  try
    for I := 0 to TestRegistry.Count - 1 do
      ClassNames.Add(TestRegistry[I].ClassName);
  finally
    ClassNames.EndUpdate;
  end;
end;

initialization
  TestRegistry := TClassList.Create;

finalization
  TestRegistry.Free;

end.



