
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxUtils.pas.                                          }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxUtils;

{$include Jvx.inc}

interface

uses
  SysUtils, Classes, TypInfo;

type

  TJvxOrdinal = type LongInt;
  TJvxMethodPointer = procedure of object;

  { TJvxNotifyComponent }

  TJvxComponentNotifyEvent = procedure(Component: TComponent;
    Operation: TOperation) of object;

  TJvxNotifyComponent = class(TComponent)
  private
    FOnNotify: TJvxComponentNotifyEvent;
  protected
    procedure Notification(Component: TComponent;
      Operation: TOperation); override;
  published
    property OnNotify: TJvxComponentNotifyEvent read FOnNotify write FOnNotify;
  end;

  { TJvxNotifyList }

  TJvxListNotifyEvent = procedure(Ptr: Pointer;
    Action: TListNotification) of object;

  TJvxNotifyList = class(TList)
  private
    FOnNotify: TJvxListNotifyEvent;
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  public
    property OnNotify: TJvxListNotifyEvent read FOnNotify write FOnNotify;
  end;

  { TJvxMethodList }

  TJvxMethodList = class(TObject)
  private
    FItems: TJvxNotifyList;
    function GetCount: Integer;
    function GetItem(Index: Integer): TJvxMethodPointer;
    procedure Notify(Ptr: Pointer; Action: TListNotification);
    procedure SetCount(Value: Integer);
    procedure SetItem(Index: Integer; Method: TJvxMethodPointer);
  public
    constructor Create;
    destructor Destroy; override;

    function Add(Method: TJvxMethodPointer): Integer;
    procedure Delete(Index: Integer);
    function IndexOf(Method: TJvxMethodPointer): Integer;
    procedure Insert(Index: Integer; Method: TJvxMethodPointer);
    procedure Pack;
    function Remove(Method: TJvxMethodPointer): Integer;

    property Count: Integer read GetCount write SetCount;
    property Items[Index: Integer]: TJvxMethodPointer read GetItem write SetItem; default;
  end;

  { TJvxCustomEvents }

  TJvxCustomEvents = class(TObject)
  private
    FEventField: ^TNotifyEvent;
    FEvents: TJvxMethodList;
    FSender: TObject;
  protected
    procedure Add(Event: TNotifyEvent);
    procedure Invoke(ASender: TObject);
    procedure InvokeEvent(Event: TNotifyEvent; var Abort: Boolean); virtual;
    procedure Remove(Event: TNotifyEvent);

    property Sender: TObject read FSender;
  public
    constructor Create(EventField: Pointer); overload;
    constructor Create; overload;
    destructor Destroy; override;
  end;

  { TJvxNotifyEvents }

  TJvxNotifyEvents = class(TJvxCustomEvents)
  public
    procedure Add(Event: TNotifyEvent);
    procedure Invoke(Sender: TObject);
    procedure Remove(Event: TNotifyEvent);
  end;

{ Utility functions }

function JvxBoolToStr(Value: Boolean): string;
procedure JvxCheckRange(TypeInfo: PTypeInfo; Value: TJvxOrdinal);
function JvxClone(Prototype: TComponent; Owner: TComponent): TComponent;
function JvxCompareMem(const A, B; Length: Integer): Boolean;
function JvxFloatToInt(Value: Extended): Integer;
function JvxStrToBool(Value: string): Boolean;

implementation

uses
  SysConst;

resourcestring
  SInvalidBoolean = '''%s'' is not a valid boolean';
  SInvalidEventFieldPointer = 'Invalid event field pointer';
  SInvalidEventHandler = 'Invalid event handler';
  SInvalidPrototype = 'Invalid prototype';
  SInvalidTypeInfo = 'Invalid type information';
  SOrdinalTypeInfoExpected = 'Ortinal type information expected';
  SValueOutOfRange = 'Ordinal value %d is out of range for type %s';

{ TJvxNotifyComponent }

procedure TJvxNotifyComponent.Notification(Component: TComponent;
  Operation: TOperation);
begin
  inherited;
  if Assigned(FOnNotify) then
    FOnNotify(Component, Operation);
end;

{ TJvxNotifyList }

procedure TJvxNotifyList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  inherited;
  if Assigned(FOnNotify) then
    FOnNotify(Ptr, Action);
end;

{ TJvxMethodList }

function TJvxMethodList.Add(Method: TJvxMethodPointer): Integer;
begin
  Result := Count;
  Insert(Result, Method);
end;

constructor TJvxMethodList.Create;
begin
  inherited Create;
  FItems := TJvxNotifyList.Create;
  FItems.OnNotify := Notify;
end;

procedure TJvxMethodList.Delete(Index: Integer);
begin
  FItems.Delete(Index);
end;

destructor TJvxMethodList.Destroy;
begin
  FItems.Free;
  inherited;
end;

function TJvxMethodList.GetCount: Integer;
begin
  Result := FItems.Count;
end;

function TJvxMethodList.GetItem(Index: Integer): TJvxMethodPointer;
var
  Item: ^TJvxMethodPointer;
begin
  Item := FItems[Index];
  if Item <> nil then
    Result := Item^
  else
    Result := nil;
end;

function TJvxMethodList.IndexOf(Method: TJvxMethodPointer): Integer;
var
  Item: TJvxMethodPointer;
begin
  for Result := 0 to Count - 1 do
  begin
    Item := Items[Result];
    if JvxCompareMem(Item, Method, SizeOf(TMethod)) then
      Exit;
  end;
  Result := -1;
end;

procedure TJvxMethodList.Insert(Index: Integer; Method: TJvxMethodPointer);
var
  Item: ^TJvxMethodPointer;
begin
  Item := nil;
  try
    if Assigned(Method) then
    begin
      New(Item);
      Item^ := Method;
    end
    else Item := nil;
    FItems.Insert(Index, Item);
  except
    if Item <> nil then
      Dispose(Item);
    raise;
  end;
end;

procedure TJvxMethodList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  if (Action = lnDeleted) and (Ptr <> nil) then
    Dispose(Ptr);
end;

procedure TJvxMethodList.Pack;
begin
  FItems.Pack;
end;

function TJvxMethodList.Remove(Method: TJvxMethodPointer): Integer;
begin
  Result := IndexOf(Method);
  Delete(Result);
end;

procedure TJvxMethodList.SetCount(Value: Integer);
begin
  FItems.Count := Value;
end;

procedure TJvxMethodList.SetItem(Index: Integer; Method: TJvxMethodPointer);
var
  Item: ^TJvxMethodPointer;
begin
  Item := FItems[Index];
  if Assigned(Method) then
  begin
    if Item = nil then
    begin
      New(Item);
      FItems[Index] := Item;
    end;
    Item^ := Method;
  end
  else FItems[Index] := nil;
end;

{ TJvxCustomEvents }

procedure TJvxCustomEvents.Add(Event: TNotifyEvent);
begin
  if not Assigned(Event) then
    raise EInvalidOperation.Create(SInvalidEventHandler);
  if FEvents = nil then
    FEvents := TJvxMethodList.Create;
  FEvents.Add(TJvxMethodPointer(Event));
end;

constructor TJvxCustomEvents.Create(EventField: Pointer);
begin
  inherited Create;
  if EventField = nil then
    raise EInvalidOperation.Create(SInvalidEventFieldPointer);
  FEventField := EventField;
end;

constructor TJvxCustomEvents.Create;
begin
  inherited;
end;

destructor TJvxCustomEvents.Destroy;
begin
  FEvents.Free;
  inherited;
end;

procedure TJvxCustomEvents.Invoke(ASender: TObject);
var
  Aborted: Boolean;
  I: Integer;
begin
  Aborted := False;
  FSender := ASender;

  if Assigned(FEventField) and Assigned(FEventField^) then
  begin
    InvokeEvent(FEventField^, Aborted);
    if Aborted then
      Exit;
  end;

  if FEvents <> nil then
    for I := 0 to FEvents.Count - 1 do
    begin
      InvokeEvent(TNotifyEvent(FEvents[I]), Aborted);
      if Aborted then
        Exit;
    end;
end;

procedure TJvxCustomEvents.InvokeEvent(Event: TNotifyEvent; var Abort: Boolean);
begin
  Event(Sender);
end;

procedure TJvxCustomEvents.Remove(Event: TNotifyEvent);
var
  Index: Integer;
begin
  if FEvents = nil then
    raise EInvalidOperation.Create(SInvalidEventHandler);
  Index := FEvents.IndexOf(TJvxMethodPointer(Event));
  if Index < 0 then
    raise EInvalidOperation.Create(SInvalidEventHandler);
  FEvents.Delete(Index);
  if FEvents.Count = 0 then
    FreeAndNil(FEvents);
end;

{ TJvxNotifyEvents }

procedure TJvxNotifyEvents.Add(Event: TNotifyEvent);
begin
  inherited;
end;

procedure TJvxNotifyEvents.Invoke(Sender: TObject);
begin
  inherited;
end;

procedure TJvxNotifyEvents.Remove(Event: TNotifyEvent);
begin
  inherited;
end;

{ Utility functions }

function JvxBoolToStr(Value: Boolean): string;
begin
  Result := GetEnumName(TypeInfo(Boolean), Ord(Value));
end;

procedure JvxCheckRange(TypeInfo: PTypeInfo; Value: TJvxOrdinal);
var
  TypeData: PTypeData;
begin
  if TypeInfo = nil then
    raise EInvalidOperation.Create(SInvalidTypeInfo);
  if not (TypeInfo.Kind in [tkInteger, tkChar, tkEnumeration, tkWChar]) then
    raise EInvalidOperation.Create(SOrdinalTypeInfoExpected);

  TypeData := GetTypeData(TypeInfo);
  if (Value < TypeData.MinValue) or (Value > TypeData.MaxValue) then
    raise ERangeError.CreateFmt(SValueOutOfRange, [Value, TypeInfo.Name]);
end;

function JvxClone(Prototype: TComponent; Owner: TComponent): TComponent;
begin
  if Prototype = nil then
    raise EInvalidOperation.Create(SInvalidPrototype);

  Result := TComponentClass(Prototype.ClassType).Create(Owner);
  try
    Result.Assign(Prototype);
  except
    Result.Free;
    raise;
  end;
end;

function JvxCompareMem(const A, B; Length: Integer): Boolean;
begin
  Result := SysUtils.CompareMem(@Byte(A), @Byte(B), Length);
end;

function JvxFloatToInt(Value: Extended): Integer;
begin
  if Frac(Value) <> 0 then
    raise EConvertError.CreateFmt(SInvalidInteger, [FloatToStr(Value)]);
  Result := Trunc(Value);
end;

function JvxStrToBool(Value: string): Boolean;
var
  EnumValue: LongInt;
begin
  EnumValue := GetEnumValue(TypeInfo(Boolean), Value);
  if (EnumValue < Ord(False)) or (EnumValue > Ord(True)) then
    raise EConvertError.CreateFmt(SInvalidBoolean, [Value]);
  Result := Boolean(EnumValue);
end;

end.
