
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxXMLClasses.pas.                                     }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxXMLClasses;

{$include Jvx.inc}

interface

uses
  Classes, Contnrs, JvxClasses, JvxStdNodes;

type

  { TJvxCustomProcessingInstruction }

  TJvxCustomProcessingInstruction = class(TJvxInstruction);

  { TJvxProcessingInstruction }

  TJvxProcessingInstruction = class(TJvxCustomProcessingInstruction)
  protected
    procedure ValidateAttribute(Attribute: TJvxNode); override;
  published
    property NodeName write SetNodeName;
    property OnNotify;
  end;

  { TJvxXMLInstruction }

  TJvxXMLInstruction = class(TJvxCustomProcessingInstruction)
  private
    FEncoding: TJvxProtectedStringAttribute;
    FStandalone: TJvxProtectedBooleanAttribute;
    FVersion: TJvxProtectedStringAttribute;
  protected
    procedure DefineAttributes(DefineAttribute: TJvxDefineNodeProc;
      Data: Pointer); override;
  public
    constructor Create(Owner: TComponent); override;
  published
    property Encoding: TJvxProtectedStringAttribute read FEncoding;
    property Standalone: TJvxProtectedBooleanAttribute read FStandalone;
    property Version: TJvxProtectedStringAttribute read FVersion;
  end;

  { TJvxXMLStyleSheetInstruction }

  TJvxXMLStyleSheetInstruction = class(TJvxCustomProcessingInstruction)
  private
    FContentType: TJvxProtectedStringAttribute;
    FHRef: TJvxProtectedStringAttribute;
  protected
    procedure DefineAttributes(DefineAttribute: TJvxDefineNodeProc;
      Data: Pointer); override;
  published
    constructor Create(Owner: TComponent); override;
    property ContentType: TJvxProtectedStringAttribute read FContentType;
    property HRef: TJvxProtectedStringAttribute read FHRef;
  end;

  { TJvxXMLDocument }

  TJvxXMLDocument = class(TJvxCompositeElement)
  private
    FXMLDeclaration: TJvxXMLInstruction;
  protected
    procedure DefineElements(DefineElement: TJvxDefineNodeProc;
      Data: Pointer); override;
    procedure ValidateElement(Element: TJvxNode); override;
  public
    function Content: string;
  published
    property XMLDeclaration: TJvxXMLInstruction read FXMLDeclaration;
  end;

  { TJvxXMLStyleSheetDocument }

  TJvxXMLStyleSheetDocument = class(TJvxXMLDocument)
  private
    FXMLStyleSheet: TJvxXMLStyleSheetInstruction;
  protected
    procedure DefineElements(DefineElement: TJvxDefineNodeProc;
      Data: Pointer); override;
  published
    property XMLStyleSheet: TJvxXMLStyleSheetInstruction read FXMLStyleSheet;
  end;

  { TJvxXMLGenerator }

  TJvxXMLGenerator = class(TJvxVisitor)
  private
    FContent: TStringStream;
    procedure CheckNodeName(Node: TJvxNode);
    procedure WriteAttributes(Node: TJvxNode);
    class function XMLString(const Source: string): string;
    class function XMLQuotedString(const Source: string): string;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Visit(Attribute: TJvxAttribute); override;
    procedure Visit(Element: TJvxElement); override;
    procedure Visit(Instruction: TJvxInstruction); override;

    function XMLFrom(Document: TJvxXMLDocument): string; overload;
    function XMLFrom(Node: TJvxNode): string; overload;
    function XMLFrom(Nodes: TJvxNodes): string; overload;
  end;

  { EJvxXMLError }

  EJvxXMLError = class(EJvxError);

implementation

uses
  SysUtils;

resourcestring
  SNodeNameIsEmpty = 'Node name is empty';

{ TJvxProcessingInstruction }

procedure TJvxProcessingInstruction.ValidateAttribute(Attribute: TJvxNode);
begin
  if Attribute is TJvxAttribute then Exit;
  inherited;
end;

{ TJvxXMLInstruction }

constructor TJvxXMLInstruction.Create(Owner: TComponent);
begin
  inherited;
  SetNodeName('xml');
  TJvxExposedNode(Encoding).NodeName := 'encoding';
  TJvxExposedNode(Standalone).NodeName := 'standalone';
  TJvxExposedNode(Version).NodeName := 'version';
  Version.AsString := '1.0';
end;

procedure TJvxXMLInstruction.DefineAttributes(
  DefineAttribute: TJvxDefineNodeProc; Data: Pointer);
begin
  inherited;
  DefineAttribute(FEncoding, 'FEncoding', TJvxProtectedStringAttribute, Data);
  DefineAttribute(FStandalone, 'FStandalone', TJvxProtectedBooleanAttribute, Data);
  DefineAttribute(FVersion, 'FVersion', TJvxProtectedStringAttribute, Data);
end;

{ TJvxXMLStyleSheetInstruction }

constructor TJvxXMLStyleSheetInstruction.Create(Owner: TComponent);
begin
  inherited;
  SetNodeName('xml-stylesheet');
  TJvxExposedNode(HRef).NodeName := 'href';
  TJvxExposedNode(ContentType).NodeName := 'type';
end;

procedure TJvxXMLStyleSheetInstruction.DefineAttributes(
  DefineAttribute: TJvxDefineNodeProc; Data: Pointer);
begin
  inherited;
  DefineAttribute(FHRef, 'FHRef', TJvxProtectedStringAttribute, Data);
  DefineAttribute(FContentType, 'FContentType', TJvxProtectedStringAttribute, Data);
end;

{ TJvxXMLDocument }

function TJvxXMLDocument.Content: string;
var
  Generator: TJvxXMLGenerator;
begin
  Generator := TJvxXMLGenerator.Create;
  try
    Result := Generator.XMLFrom(Self);
  finally
    Generator.Free;
  end;
end;

procedure TJvxXMLDocument.DefineElements(
  DefineElement: TJvxDefineNodeProc; Data: Pointer);
begin
  inherited;
  DefineElement(FXMLDeclaration, 'FXMLDeclaration', TJvxXMLInstruction, Data);
end;

procedure TJvxXMLDocument.ValidateElement(Element: TJvxNode);
begin
  if Element is TJvxElement then Exit;
  inherited;
end;

{ TJvxXMLStyleSheetDocument }

procedure TJvxXMLStyleSheetDocument.DefineElements(
  DefineElement: TJvxDefineNodeProc; Data: Pointer);
begin
  inherited;
  DefineElement(FXMLStyleSheet, 'FXMLStyleSheet', TJvxXMLStyleSheetInstruction, Data);
end;

{ TJvxXMLGenerator }

procedure TJvxXMLGenerator.CheckNodeName(Node: TJvxNode);
begin
  if Node.NodeName = EmptyStr then
    raise EJvxXMLError.Create(SNodeNameIsEmpty);
end;

constructor TJvxXMLGenerator.Create;
begin
  inherited;
  FContent := TStringStream.Create(EmptyStr);
end;

destructor TJvxXMLGenerator.Destroy;
begin
  FContent.Free;
  inherited;
end;

procedure TJvxXMLGenerator.Visit(Attribute: TJvxAttribute);
begin
  CheckNodeName(Attribute);
  if Attribute.Assigned then
  begin
    FContent.WriteString(' ' + Attribute.NodeName + '=');
    FContent.WriteString(XMLQuotedString(Attribute.AsString));
  end;
end;

procedure TJvxXMLGenerator.Visit(Element: TJvxElement);

  function NeedClosingTag: Boolean;
  begin
    Result := Element.HasElements or Element.Assigned;
  end;

  procedure WriteClosingTag;
  begin
    if NeedClosingTag then
      FContent.WriteString('</' + Element.NodeName + '>');
  end;

  procedure WriteElements;
  begin
    if Element.HasElements then
      Element.Elements.Accept(Self);
  end;

  procedure WriteOpeningTag;
  begin
    FContent.WriteString('<' + Element.NodeName);
    WriteAttributes(Element);
    if not NeedClosingTag then
      FContent.WriteString('/');
    FContent.WriteString('>');
  end;

  procedure WriteValue;
  begin
    if Element.Assigned then
      FContent.WriteString(XMLString(Element.AsString));
  end;

begin
  CheckNodeName(Element);
  WriteOpeningTag;
  WriteValue;
  WriteElements;
  WriteClosingTag;
end;

procedure TJvxXMLGenerator.Visit(Instruction: TJvxInstruction);
begin
  CheckNodeName(Instruction);
  FContent.WriteString('<?' + Instruction.NodeName);
  WriteAttributes(Instruction);
  FContent.WriteString('?>');
end;

function TJvxXMLGenerator.XMLFrom(Document: TJvxXMLDocument): string;
begin
  Result := XMLFrom(Document.Elements);
end;

function TJvxXMLGenerator.XMLFrom(Node: TJvxNode): string;
begin
  FContent.Seek(0, soFromBeginning);
  Node.Accept(Self);
  Result := FContent.DataString;
end;

procedure TJvxXMLGenerator.WriteAttributes(Node: TJvxNode);
begin
  if Node.HasAttributes then
    Node.Attributes.Accept(Self);
end;

function TJvxXMLGenerator.XMLFrom(Nodes: TJvxNodes): string;
begin
  FContent.Seek(0, soFromBeginning);
  Nodes.Accept(Self);
  Result := FContent.DataString;
end;

class function TJvxXMLGenerator.XMLQuotedString(const Source: string): string;
begin
  Result := StringReplace(XMLString(Source), '''', '&apos;', [rfReplaceAll]);
  Result := '''' + Result + '''';
end;

class function TJvxXMLGenerator.XMLString(const Source: string): string;
begin
  Result := StringReplace(Source, '&', '&amp;', [rfReplaceAll]);
  Result := StringReplace(Result, '<', '&lt;', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '&gt;', [rfReplaceAll]);
end;

end.

