
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxDebugMemTest.pas.                                   }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxDebugMemTest;

{$include Jvx.inc}

interface

uses
  Classes, JvxTestClasses;

type

  { TJvxDebugMemTest }

  TJvxDebugMemTest = class(TJvxTestSuite)
  private
    procedure CheckMemory(Buffer: PChar; Size: Integer; Value: Char);
  public
    constructor Create(Owner: TComponent); override;
  published
    procedure TestAllocMemCount;
    procedure TestAllocMemSize;
    procedure TestGetFreeMem;
    procedure TestHeapStatus;
    procedure TestReallocMem;
    procedure TestValidPointer;
    procedure TestValidObject;
    procedure TestValidMethod;
  end;

implementation

uses
  SysUtils, JvxMemoryManager, JvxDebugMem;

const
  JvxTestBufferSize = 1024;

{ TJvxDebugMemTest }

procedure TJvxDebugMemTest.CheckMemory(Buffer: PChar; Size: Integer;
  Value: Char);
var
  I: Integer;
begin
  for I := 0 to Size - 1 do
    if not (Buffer[I] = Value) then
      // Use if statement to avoid allocating temporary strings 
      Assert(False, 'Buffer[' + IntToStr(I) + '] = #$' + IntToHex(Ord(Buffer[I]), 2));
end;

constructor TJvxDebugMemTest.Create(Owner: TComponent);
begin
  inherited;
  AddTest(TestAllocMemCount);
  AddTest(TestAllocMemSize);
  AddTest(TestGetFreeMem);
  AddTest(TestHeapStatus);
  AddTest(TestReallocMem);
  AddTest(TestValidPointer);
  AddTest(TestValidObject);
  AddTest(TestValidMethod);
end;

procedure TJvxDebugMemTest.TestAllocMemCount;
var
  MemCount: Integer;
  P: Pointer;
begin
  MemCount := JvxDebugAllocMemCount;
  Assert(MemCount > 0);

  GetMem(P, JvxTestBufferSize);
  try
    Assert(JvxDebugAllocMemCount = (MemCount + 1));
  finally
    FreeMem(P);
  end;

  Assert(JvxDebugAllocMemCount = MemCount);
end;

procedure TJvxDebugMemTest.TestAllocMemSize;
var
  MemSize: Integer;
  P: Pointer;
begin
  MemSize := JvxDebugAllocMemSize;
  Assert(MemSize > 0);

  GetMem(P, JvxTestBufferSize);
  try
    Assert(JvxDebugAllocMemSize = (MemSize + JvxTestBufferSize));
  finally
    FreeMem(P);
  end;

  Assert(JvxDebugAllocMemSize = MemSize);
end;

procedure TJvxDebugMemTest.TestGetFreeMem;
var
  Buffer: PChar;
begin
  // Test that GetMem fills new blocks with garbage byte
  GetMem(Buffer, JvxTestBufferSize);
  try
    CheckMemory(Buffer, JvxTestBufferSize, JvxGetMemGarbageChar);
  finally
    FreeMem(Buffer);
  end;

  // Test that FreeMem fills released blocks with garbage byte.
  // Note that first 2 DWORDS of a freed block contain memory manager's data, so
  // they should not be checked.
  CheckMemory(Buffer + SizeOf(Pointer) * 2,
    JvxTestBufferSize - SizeOf(Pointer) * 2 - SizeOf(Integer),
    JvxFreeMemGarbageChar);
end;

procedure TJvxDebugMemTest.TestHeapStatus;
begin
  Assert(JvxDebugHeapStatus.HeapErrorCode = 0);
end;

procedure TJvxDebugMemTest.TestReallocMem;
var
  Buffer: PChar;
  OldBuffer: PChar;
begin
  GetMem(Buffer, JvxTestBufferSize);
  try
    // Test that ReallocMem always moves a memory block to a new location when
    // it is shrinked
    OldBuffer := Buffer;
    ReallocMem(Buffer, JvxTestBufferSize div 2);
    Assert(Buffer <> OldBuffer);

    // Test that the memory is filled with garbage in the old location
    CheckMemory(OldBuffer + SizeOf(Pointer) * 2,
      JvxTestBufferSize - SizeOf(Pointer) * 2 - SizeOf(Integer),
      JvxFreeMemGarbageChar);

    FillChar(Buffer^, JvxTestBufferSize div 2, 'A');

    // Test that ReallocMem always moves a memory block to a new location when
    // it is expanded
    OldBuffer := Buffer;
    ReallocMem(Buffer, JvxTestBufferSize * 2);
    Assert(Buffer <> OldBuffer);

    // Test that ReallocMem preserves the content of a memory block when moving
    CheckMemory(Buffer, JvxTestBufferSize div 2, 'A');

    // Test that the remaining space is filled with garbage
    CheckMemory(Buffer + JvxTestBufferSize div 2, (JvxTestBufferSize div 2) * 3,
      JvxGetMemGarbageChar);
  finally
    FreeMem(Buffer);
  end;
end;

procedure TJvxDebugMemTest.TestValidMethod;
var
  Obj: TComponent;
  Method: procedure of object;
begin
  // Test with nil
  Obj := nil;
  Method := Obj.Destroying;
  Assert(not JvxValidMethod(TMethod(Method)));

  Obj := TComponent.Create(nil);
  try
    Method := Obj.Destroying;

    // Test with valid object
    Assert(JvxValidMethod(TMethod(Method)));
  finally
    Obj.Free;
  end;

  // Test with freed object
  Assert(not JvxValidMethod(TMethod(Method)));
end;

procedure TJvxDebugMemTest.TestValidObject;
var
  Obj: TComponent;
begin
  // Test with nil
  Obj := nil;
  Assert(not JvxValidObject(Obj));
  Assert(not JvxValidObject(Obj, TComponent));

  Obj := TComponent.Create(nil);
  try
    // Test exact match
    Assert(JvxValidObject(Obj));
    Assert(JvxValidObject(Obj, TComponent));

    // Test pointer pointing into the middle of the block
    Assert(not JvxValidObject(TObject(PChar(Obj)+ 5)));
    Assert(not JvxValidObject(TObject(PChar(Obj)+ 5), TComponent));

    // Test with different class
    Assert(JvxValidObject(Obj, TPersistent));
    Assert(not JvxValidObject(Obj, TJvxTestCase));
  finally
    Obj.Free;
  end;

  // Test with freed object
  Assert(not JvxValidObject(Obj));
  Assert(not JvxValidObject(Obj, TComponent));
end;

procedure TJvxDebugMemTest.TestValidPointer;
var
  Buffer: PChar;
begin
  // Test with nil
  Buffer := nil;
  Assert(not JvxValidPointer(Buffer));
  Assert(not JvxValidPointer(Buffer, 0));

  GetMem(Buffer, JvxTestBufferSize);
  try
    // Test exact match
    Assert(JvxValidPointer(Buffer));
    Assert(JvxValidPointer(Buffer, JvxTestBufferSize));

    // Test pointer pointing into the middle of the block
    Assert(not JvxValidPointer(Buffer + 5));
    Assert(not JvxValidPointer(Buffer + 5, JvxTestBufferSize));

    // Test that block is invalid if it's actual size is smaller than expected
    Assert(not JvxValidPointer(Buffer, JvxTestBufferSize * 2));

    // Test that block is valid even if it's actual size is bigger than expected
    Assert(JvxValidPointer(Buffer, JvxTestBufferSize div 2));
  finally
    FreeMem(Buffer);
  end;

  // Test with freed block
  Assert(not JvxValidPointer(Buffer));
  Assert(not JvxValidPointer(Buffer, JvxTestBufferSize));
end;

initialization
  JvxRegisterTestClass(TJvxDebugMemTest);
end.
