
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxUtilsTest.pas.                                      }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxUtilsTest;

interface

uses
  Classes, JvxTestClasses, JvxUtils;

type
  { TNotifyComponentTest }

  TNotifyComponentTest = class(TJvxTestSuite)
  private
    FNotifier: TJvxNotifyComponent;
    FNotificationComponent: TComponent;
    FNotificationOperation: TOperation;
    procedure OnNotify(Component: TComponent;
      Operation: TOperation);
  protected
    procedure InternalCreateFixture; override;
    procedure InternalDestroyFixture; override;
  public
    constructor Create(Owner: TComponent); override;
  published
    procedure TestInsertNotification;
    procedure TestRemoveNotification;
  end;

  { TNotifyListTest }

  TNotifyListTest = class(TJvxTestSuite)
  private
    FList: TJvxNotifyList;
    FNotifyAction: TListNotification;
    FNotifyPtr: Pointer;
    procedure Notify(Ptr: Pointer; Action: TListNotification);
  protected
    procedure InternalCreateFixture; override;
    procedure InternalDestroyFixture; override;
  public
    constructor Create(Owner: TComponent); override;
  published
    procedure TestOnNotify;
  end;

  { TMethodListTest }

  TMethodListTest = class(TJvxTestSuite)
  private
    FList: TJvxMethodList;
  protected
    procedure InternalCreateFixture; override;
    procedure InternalDestroyFixture; override;
  public
    constructor Create(Owner: TComponent); override;
  published
    procedure TestAdd;
    procedure TestCount;
    procedure TestDelete;
    procedure TestGetItem;
    procedure TestIndexOf;
    procedure TestInsert;
    procedure TestPack;
    procedure TestRemove;
    procedure TestSetItem;
  end;

  { TTestEvents }

  TTestEvent = procedure(Sender: TObject; var Abort: Boolean) of object;
  TTestEvents = class(TJvxCustomEvents)
  protected
    procedure InvokeEvent(AEvent: TNotifyEvent; var Abort: Boolean); override;
  public
    procedure Add(AEvent: TTestEvent);
    procedure Remove(AEvent: TTestEvent);
  end;

  { TCustomEventsTest }

  TCustomEventsTest = class(TJvxTestSuite)
  private
    FFirstEventSender: TObject;
    FFirstHandlerCounter: Integer;
    FEvents: TTestEvents;
    FEvent: TTestEvent;
    FSecondEventSender: TObject;
    FSecondHandlerCounter: Integer;
    procedure FirstHandler(Sender: TObject; var Abort: Boolean);
    procedure SecondHandler(Sender: TObject; var Abort: Boolean);
  protected
    procedure InternalCreateFixture; override;
    procedure InternalDestroyFixture; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    procedure TestAdd;
    procedure TestEvent;
    procedure TestInvoke;
    procedure TestRemove;
  end;

implementation

uses
  SysUtils;

{ TTestHandler }

type
  TTestHandler = class(TComponent)
  public
    procedure OnEvent(Sender: TObject);
  end;

procedure TTestHandler.OnEvent(Sender: TObject);
begin
end;

{ TNotifyComponentTest }

constructor TNotifyComponentTest.Create(Owner: TComponent);
begin
  inherited;
  AddTest(TestInsertNotification);
  AddTest(TestRemoveNotification);
end;

procedure TNotifyComponentTest.InternalCreateFixture;
begin
  inherited;
  FNotifier := TJvxNotifyComponent.Create(Self);
  FNotifier.OnNotify := Self.OnNotify;
  FNotificationComponent := nil;
  FNotificationOperation := opRemove;
end;

procedure TNotifyComponentTest.InternalDestroyFixture;
begin
  FreeAndNil(FNotifier);
  inherited;
end;

procedure TNotifyComponentTest.OnNotify(Component: TComponent;
  Operation: TOperation);
begin
  FNotificationComponent := Component;
  FNotificationOperation := Operation;
end;

procedure TNotifyComponentTest.TestInsertNotification;
var
  Component: TComponent;
begin
  Component := TComponent.Create(FNotifier);
  try
    Assert(FNotificationComponent = Component);
    Assert(FNotificationOperation = opInsert);
  finally
    Component.Free;
  end;
end;

procedure TNotifyComponentTest.TestRemoveNotification;
var
  Component: TComponent;
begin
  Component := TComponent.Create(FNotifier);
  try
    FNotificationComponent := nil;
    FNotificationOperation := opInsert;
  finally
    Component.Free;
  end;

  Assert(FNotificationComponent = Component);
  Assert(FNotificationOperation = opRemove);
end;

{ TNotifyListTest }

constructor TNotifyListTest.Create(Owner: TComponent);
begin
  inherited;
  AddTest(TestOnNotify);
end;

procedure TNotifyListTest.InternalCreateFixture;
begin
  inherited;
  FList := TJvxNotifyList.Create;
  FList.OnNotify := Notify;
  FNotifyAction := Low(TListNotification);
  FNotifyPtr := nil;
end;

procedure TNotifyListTest.InternalDestroyFixture;
begin
  FList.Free;
  inherited;
end;

procedure TNotifyListTest.Notify(Ptr: Pointer; Action: TListNotification);
begin
  FNotifyPtr := Ptr;
  FNotifyAction := Action;
end;

procedure TNotifyListTest.TestOnNotify;
begin
  FList.Add(Pointer($FFFFFFFF));
  Assert(FNotifyPtr = Pointer($FFFFFFFF));
  Assert(FNotifyAction = lnAdded);

  FList.Add(Pointer($AAAAAAAA));
  Assert(FNotifyPtr = Pointer($AAAAAAAA));
  Assert(FNotifyAction = lnAdded);

  FList.Remove(Pointer($FFFFFFFF));
  Assert(FNotifyPtr = Pointer($FFFFFFFF));
  Assert(FNotifyAction = lnDeleted);

  FList.Extract(Pointer($AAAAAAAA));
  Assert(FNotifyPtr = Pointer($AAAAAAAA));
  Assert(FNotifyAction = lnExtracted);
end;

{ TMethodListTest }

constructor TMethodListTest.Create(Owner: TComponent);
begin
  inherited;
  AddTest(TestAdd);
  AddTest(TestCount);
  AddTest(TestDelete);
  AddTest(TestGetItem);
  AddTest(TestIndexOf);
  AddTest(TestInsert);
  AddTest(TestPack);
  AddTest(TestRemove);
  AddTest(TestSetItem);
end;

procedure TMethodListTest.InternalCreateFixture;
begin
  inherited;
  FList := TJvxMethodList.Create;
end;

procedure TMethodListTest.InternalDestroyFixture;
begin
  FreeAndNil(FList);
  inherited;
end;

procedure TMethodListTest.TestAdd;
begin
  Assert(FList.Count = 0);
  Assert(FList.Add(TestAdd) = 0);
  Assert(FList.Count = 1);
end;

procedure TMethodListTest.TestCount;
begin
  // Test GetCount
  Assert(FList.Count = 0);
  FList.Add(nil);
  Assert(FList.Count = 1);

  // Test SetCount
  FList.Count := 2;
  Assert(not Assigned(FList.Items[1]));

  FList.Count := 0;
  try
    FList.Items[0];
    Assert(EListError);
  except
    Assert(EListError);
  end;
end;

procedure TMethodListTest.TestDelete;
begin
  // Test that Delete actually works
  Assert(FList.Count = 0);
  FList.Add(nil);
  Assert(FList.Count = 1);
  FList.Delete(0);
  Assert(FList.Count = 0);

  // Test that Delete raises an exception if the index is invalid
  try
    FList.Delete(0);
    Assert(EListError);
  except
    Assert(EListError);
  end;
end;

procedure TMethodListTest.TestGetItem;
begin
  Assert(FList.Count = 0);

  FList.Add(nil);
  Assert(not Assigned(FList.Items[0]));

  FList.Add(TestGetItem);
  Assert(Assigned(FList.Items[1]));

  try
    FList.Items[-1];
    Assert(EListError);
  except
    Assert(EListError);
  end;

  try
    FList.Items[FList.Count];
    Assert(EListError);
  except
    Assert(EListError);
  end;

  // Test that after increasing TJvxMethodList.Count, new items are nils
  FList.Count := FList.Count + 1;
  Assert(not Assigned(FList.Items[FList.Count - 1]));
end;

procedure TMethodListTest.TestIndexOf;
begin
  Assert(FList.IndexOf(TestIndexOf) < 0);
  FList.Add(TestIndexOf);
  Assert(FList.IndexOf(TestIndexOf) >= 0);
end;

procedure TMethodListTest.TestInsert;
begin
  Assert(FList.Count = 0);

  FList.Add(nil);
  FList.Add(nil);
  FList.Add(nil);

  try
    FList.Insert(-1, TestInsert);
    Assert(EListError);
  except
    Assert(EListError);
  end;

  try
    FList.Insert(FList.Count + 1, TestInsert);
    Assert(EListError);
  except
    Assert(EListError);
  end;

  FList.Insert(1, TestInsert);
  Assert(FList.IndexOf(TestInsert) = 1);
end;

procedure TMethodListTest.TestPack;
begin
  Assert(FList.Count = 0);
  FList.Add(nil);
  Assert(FList.Count = 1);
  FList.Pack;
  Assert(FList.Count = 0);
end;

procedure TMethodListTest.TestRemove;
begin
  Assert(FList.Count = 0);
  FList.Add(TestRemove);
  Assert(FList.IndexOf(TestRemove) = 0);
  Assert(FList.Remove(TestRemove) = 0);
  Assert(FList.Count = 0);
end;

procedure TMethodListTest.TestSetItem;
begin
  Assert(FList.Count = 0);
  FList.Add(nil);
  FList.Items[0] := TestSetItem;
  Assert(FList.IndexOf(TestSetItem) = 0);
  FList.Delete(0);

  // Test that after increasing count we still can assign to an item
  FList.Count := 1;
  FList.Items[0] := TestSetItem;
  Assert(FList.IndexOf(TestSetItem) = 0);

  // Test that after assigning nil there is no memory leak
  FList.Items[0] := nil;
  Assert(not Assigned(FList.Items[0]));
  FList.Items[0] := TestSetItem;
  Assert(FList.IndexOf(TestSetItem) = 0);
end;

{ TCustomEventsTest }

constructor TCustomEventsTest.Create(AOwner: TComponent);
begin
  inherited;
  AddTest(TestAdd);
  AddTest(TestEvent);
  AddTest(TestInvoke);
  AddTest(TestRemove);
end;

procedure TCustomEventsTest.InternalCreateFixture;
begin
  inherited;
  FEvents := TTestEvents.Create(@@FEvent);
  FEvent := nil;
  FFirstEventSender := nil;
  FFirstHandlerCounter := 0;
  FSecondEventSender := nil;
  FSecondHandlerCounter := 0;
end;

procedure TCustomEventsTest.InternalDestroyFixture;
begin
  FreeAndNil(FEvents);
  inherited;
end;

procedure TCustomEventsTest.FirstHandler(Sender: TObject; var Abort: Boolean);
begin
  FFirstEventSender := Sender;
  Inc(FFirstHandlerCounter);
end;

procedure TCustomEventsTest.SecondHandler(Sender: TObject; var Abort: Boolean);
begin
  FSecondEventSender := Sender;
  Inc(FSecondHandlerCounter);
  Abort := True;
end;

procedure TCustomEventsTest.TestAdd;
begin
  // Test that Add raises an exception when invalid event handler is added
  try
    FEvents.Add(nil);
    Assert(EInvalidOperation);
  except
    Assert(EInvalidOperation);
  end;

  // Test that Add adds event handler to the list
  Assert(FFirstHandlerCounter = 0);
  FEvents.Add(FirstHandler);
  FEvents.Invoke(Self);
  Assert(FFirstHandlerCounter = 1);

  // Test that Add allows duplicates
  FEvents.Add(FirstHandler);
  FEvents.Invoke(Self);
  Assert(FFirstHandlerCounter = 3);

  // Test that adding another event handler is successful
  FEvents.Add(SecondHandler);
  FEvents.Invoke(Self);
  Assert(FSecondHandlerCounter = 1);
end;

procedure TCustomEventsTest.TestEvent;
begin
  FEvent := FirstHandler;
  FEvents.Invoke(Self);
  Assert(FFirstHandlerCounter = 1);
  Assert(FFirstEventSender = Self);
end;

procedure TCustomEventsTest.TestInvoke;
begin
  // Test invocation of event in the list
  FEvents.Add(FirstHandler);
  FEvents.Add(FirstHandler);
  FEvents.Invoke(Self);
  Assert(FFirstHandlerCounter = 2);
  Assert(FFirstEventSender = Self);

  FEvents.Remove(FirstHandler);
  FEvents.Remove(FirstHandler);
  FFirstHandlerCounter := 0;
  FFirstEventSender := nil;

  // Test invocation of a singlecast event
  FEvent := FirstHandler;
  FEvents.Invoke(Self);
  Assert(FFirstHandlerCounter = 1);
  Assert(FFirstEventSender = Self);

  FEvent := nil;
  FFirstHandlerCounter := 0;
  FFirstEventSender := nil;

  // Test abortion of event invocation (singlecast event is invoked first).
  FEvents.Add(SecondHandler);
  FEvents.Add(FirstHandler);
  FEvent := FirstHandler;
  FEvents.Invoke(Self);
  Assert(FSecondHandlerCounter = 1);
  Assert(FFirstHandlerCounter = 1);
end;

procedure TCustomEventsTest.TestRemove;
begin
  // Test that removing handler twice causes an exception
  FEvents.Add(FirstHandler);
  FEvents.Remove(FirstHandler);
  try
    FEvents.Remove(FirstHandler);
    Assert(EInvalidOperation);
  except
    Assert(EInvalidOperation);
  end;

  // Test that the first removal was successful
  FEvents.Invoke(Self);
  Assert(FFirstHandlerCounter = 0);
end;

{ TTestEvents }

procedure TTestEvents.Add(AEvent: TTestEvent);
begin
  inherited Add(TNotifyEvent(AEvent));
end;

procedure TTestEvents.InvokeEvent(AEvent: TNotifyEvent; var Abort: Boolean);
begin
  TTestEvent(AEvent)(Sender, Abort);
end;

procedure TTestEvents.Remove(AEvent: TTestEvent);
begin
  inherited Remove(TNotifyEvent(AEvent));
end;

initialization
  JvxRegisterTestClass(TNotifyComponentTest);
  JvxRegisterTestClass(TNotifyListTest);
  JvxRegisterTestClass(TMethodListTest);
  JvxRegisterTestClass(TCustomEventsTest);
end.
