
{******************************************************************************}
{                                                                              }
{  The contents of this file are subject to the Mozilla Public License         }
{  (the "License"); you may not use this file except in compliance with        }
{  the License. You may obtain a copy of the License at                        }
{  http://www.mozilla.org/MPL/                                                 }
{                                                                              }
{  Software distributed under the License is distributed on an "AS IS"         }
{  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See         }
{  the License for the specific language governing rights and                  }
{  limitations under the License.                                              }
{                                                                              }
{  The Original Code is JvxXMLClassesTest.pas.                                 }
{                                                                              }
{  The Initial Developer of the Original Code is Oleg Sych <osych@hotmail.com> }
{  Portions created by Oleg Sych are Copyright (C) 2000 Oleg Sych.             }
{  All Rights Reserved.                                                        }
{                                                                              }
{  You may retrieve the latest version of this file at XVCL home page, located }
{  at http://go.to/xvcl                                                        }
{                                                                              }
{  Contributor(s):                                                             }
{                                                                              }
{******************************************************************************}

unit JvxXMLClassesTest;

interface

uses
  Classes, JvxTestClasses, JvxXMLClasses, JvxStdNodes;

type

  { TXMLGeneratorTest }

  TXMLGeneratorTest = class(TJvxTestSuite)
  private
    FAttribute: TJvxStringAttribute;
    FElementA: TJvxCompositeTextElement;
    FElementB1: TJvxCompositeTextElement;
    FElementB2: TJvxCompositeTextElement;
    FElementC1: TJvxStringElement;
    FElementC2: TJvxStringElement;
    FElementC3: TJvxStringElement;
    FGenerator: TJvxXMLGenerator;
    FInstruction: TJvxProcessingInstruction;
  protected
    procedure InternalCreateFixture; override;
    procedure InternalDestroyFixture; override;
  public
    constructor Create(Owner: TComponent); override;
  published
    procedure TestXMLErrorEmptyNodeName;
    procedure TestXMLFromElementHierarchy;
    procedure TestXMLFromElementWithAttributes;
    procedure TestXMLFromElementWithValue;
    procedure TestXMLFromEmptyElement;
    procedure TestXMLFromNodeList;
    procedure TestXMLFromProcessingInstruction;
  end;

implementation

uses
  SysUtils;

{ TXMLGeneratorTest }

constructor TXMLGeneratorTest.Create(Owner: TComponent);
begin
  inherited;
  AddTest(TestXMLErrorEmptyNodeName);
  AddTest(TestXMLFromElementHierarchy);
  AddTest(TestXMLFromElementWithAttributes);
  AddTest(TestXMLFromElementWithValue);
  AddTest(TestXMLFromEmptyElement);
  AddTest(TestXMLFromNodeList);
  AddTest(TestXMLFromProcessingInstruction);
end;

procedure TXMLGeneratorTest.InternalCreateFixture;
begin
  inherited;

  FAttribute := TJvxStringAttribute.Create(nil);
  FAttribute.NodeName := 'attr';    
  FAttribute.AsString := '&test''<>';

  FElementA := TJvxCompositeTextElement.Create(nil);
  FElementA.NodeName := 'a';

  FElementB1 := TJvxCompositeTextElement.Create(nil);
  FElementB1.NodeName := 'b';

  FElementB2 := TJvxCompositeTextElement.Create(nil);
  FElementB2.NodeName := 'b';

  FElementC1 := TJvxStringElement.Create(nil);
  FElementC1.NodeName := 'c';

  FElementC2 := TJvxStringElement.Create(nil);
  FElementC2.NodeName := 'c';

  FElementC3 := TJvxStringElement.Create(nil);
  FElementC3.NodeName := 'c';

  FInstruction := TJvxProcessingInstruction.Create(nil);
  FInstruction.NodeName := 'xml';

  FGenerator := TJvxXMLGenerator.Create;
end;

procedure TXMLGeneratorTest.InternalDestroyFixture;
begin
  FreeAndNil(FGenerator);
  FreeAndNil(FAttribute);
  FreeAndNil(FInstruction);
  FreeAndNil(FElementC1);
  FreeAndNil(FElementC2);
  FreeAndNil(FElementC3);
  FreeAndNil(FElementB1);
  FreeAndNil(FElementB2);
  FreeAndNil(FElementA);
  inherited;
end;

procedure TXMLGeneratorTest.TestXMLErrorEmptyNodeName;
begin
  // Test exception when element's node name is empty
  FElementA.NodeName := EmptyStr;
  try
    FGenerator.XMLFrom(FElementA);
    Assert(EJvxXMLError);
  except
    Assert(EJvxXMLError);
  end;

  // Test exception when attribute's node name is empty
  FAttribute.NodeName := EmptyStr;
  FAttribute.Parent := FElementB1;
  try
    FGenerator.XMLFrom(FElementB1);
    Assert(EJvxXMLError);
  except
    Assert(EJvxXMLError);
  end;
end;

procedure TXMLGeneratorTest.TestXMLFromElementHierarchy;
const
  MatchPattern = '<a><b><c/><c/></b><b><c/></b></a>';
begin
  FElementB1.Parent := FElementA;
  FElementC1.Parent := FElementB1;
  FElementC2.Parent := FElementB1;
  FElementB2.Parent := FElementA;
  FElementC3.Parent := FElementB2;
  Assert(FGenerator.XMLFrom(FElementA) = MatchPattern);
end;

procedure TXMLGeneratorTest.TestXMLFromElementWithAttributes;
const
  MatchPattern = '<a attr=''&amp;test&apos;&lt;&gt;''/>';
begin
  FAttribute.Parent := FElementA;
  Assert(FGenerator.XMLFrom(FElementA) = MatchPattern);
end;

procedure TXMLGeneratorTest.TestXMLFromElementWithValue;
const
  MatchPattern = '<c>test</c>';
begin
  FElementC1.AsString := 'test';
  Assert(FGenerator.XMLFrom(FElementC1) = MatchPattern);
end;

procedure TXMLGeneratorTest.TestXMLFromEmptyElement;
const
  MatchPattern = '<a/>';
begin
  Assert(FGenerator.XMLFrom(FElementA) = MatchPattern);
end;

procedure TXMLGeneratorTest.TestXMLFromNodeList;
const
  MatchPattern = '<?xml attr=''&amp;test&apos;&lt;&gt;''?><b><c/></b>';
begin
  FInstruction.Parent := FElementA;
  FAttribute.Parent := FInstruction;
  FElementB1.Parent := FElementA;
  FElementC1.Parent := FElementB1;
  Assert(FGenerator.XMLFrom(FElementA.Elements) = MatchPattern);
end;

procedure TXMLGeneratorTest.TestXMLFromProcessingInstruction;
const
  MatchPattern = '<?xml attr=''&amp;test&apos;&lt;&gt;''?>';
begin
  FAttribute.Parent := FInstruction;
  Assert(FGenerator.XMLFrom(FInstruction) = MatchPattern);
end;

initialization
  JvxRegisterTestClass(TXMLGeneratorTest);

end.
